@@ -1,6 +1,6 @@
 /*************************************************************************
  *
- * $Id: trionan.c,v 1.12 2003/08/06 04:43:53 wbrack Exp $
+ * $Id: trionan.c,v 1.13 2003/08/13 11:09:42 wbrack Exp $
  *
  * Copyright (C) 2001 Bjorn Reese <breese@users.sourceforge.net>
  *
@@ -51,7 +51,11 @@
 # include <signal.h>
 #endif
 #if defined(TRIO_COMPILER_DECC)
-# include <fp_class.h>
+#  if defined(__linux__)
+#   include <cpml.h>
+#  else
+#   include <fp_class.h>
+#  endif
 #endif
 #include <assert.h>
 
@@ -108,7 +112,7 @@
  * Constants
  */
 
-static TRIO_CONST char rcsid[] = "@(#)$Id: trionan.c,v 1.12 2003/08/06 04:43:53 wbrack Exp $";
+static TRIO_CONST char rcsid[] = "@(#)$Id: trionan.c,v 1.13 2003/08/13 11:09:42 wbrack Exp $";
 
 #if defined(USE_IEEE_754)
 
@@ -435,7 +439,7 @@ trio_isinf
 TRIO_ARGS1((number),
 	   double number)
 {
-#if defined(TRIO_COMPILER_DECC)
+#if defined(TRIO_COMPILER_DECC) && !defined(__linux__)
   /*
    * DECC has an isinf() macro, but it works differently than that
    * of C99, so we use the fp_class() function instead.
